@@ -4,7 +4,6 @@ This module provides Runner class to run behave feature files (or model elements
 """
 
 from __future__ import absolute_import, print_function, with_statement
-
 import contextlib
 import os.path
 import sys
@@ -13,14 +12,28 @@ import weakref
 
 import six
 
-from behave._types import ExceptionUtil
-from behave.capture import CaptureController
-from behave.configuration import ConfigError
+from behave._types import (
+    require_type,
+    require_not_type,
+    require_callable,
+    require_not_none,
+)
+from behave.api.runner import ITestRunner
+from behave.exception_util import ExceptionUtil
+from behave.capture import (
+    CaptureController,
+    CaptureSinkAsCollector,
+    ManyCaptured,
+    any_hook, capture_output_to_sink,
+)
+from behave.exception import ConfigError
 from behave.formatter._registry import make_formatters
-from behave.runner_util import \
-    collect_feature_locations, parse_features, \
+from behave.runner_util import (
+    collect_feature_locations, parse_features,
     exec_file, load_step_modules, PathManager
+)
 from behave.step_registry import registry as the_step_registry
+from enum import Enum
 
 if six.PY2:
     # -- USE PYTHON3 BACKPORT: With unicode traceback support.
@@ -29,22 +42,29 @@ else:
     import traceback
 
 
-class CleanupError(RuntimeError):
-    pass
-
 
 class ContextMaskWarning(UserWarning):
     """Raised if a context variable is being overwritten in some situations.
 
     If the variable was originally set by user code then this will be raised if
-    *behave* overwites the value.
+    *behave* overwrites the value.
 
     If the variable was originally set by *behave* then this will be raised if
-    user code overwites the value.
+    user code overwrites the value.
     """
     pass
 
 
+class ContextMode(Enum):
+    """Used to distinguish between the two usage modes while using the context:
+
+    * BEHAVE: Indicates "behave" (internal) mode
+    * USER: Indicates "user" mode (in steps, hooks, fixtures, ...)
+    """
+    BEHAVE = 1
+    USER = 2
+
+
 class Context(object):
     """Hold contextual information during the running of tests.
 
@@ -111,30 +131,20 @@ class Context(object):
       :class:`~behave.model.Row` that is active for the current scenario. It is
       present mostly for debugging, but may be useful otherwise.
 
-    .. attribute:: log_capture
+    .. attribute:: captured
 
-      If logging capture is enabled then this attribute contains the captured
-      logging as an instance of :class:`~behave.log_capture.LoggingCapture`.
-      It is not present if logging is not being captured.
+        If any output capture is enabled, provides access to a
+        :class:`~behave.capture.Captured` object that contains a snapshot
+        of all captured data (stdout/stderr/log).
 
-    .. attribute:: stdout_capture
+        .. versionadded:: 1.3.0
 
-      If stdout capture is enabled then this attribute contains the captured
-      output as a StringIO instance. It is not present if stdout is not being
-      captured.
-
-    .. attribute:: stderr_capture
-
-      If stderr capture is enabled then this attribute contains the captured
-      output as a StringIO instance. It is not present if stderr is not being
-      captured.
-
-    If an attempt made by user code to overwrite one of these variables, or
-    indeed by *behave* to overwite a user-set variable, then a
-    :class:`behave.runner.ContextMaskWarning` warning will be raised.
+    A :class:`behave.runner.ContextMaskWarning` warning will be raised if user
+    code attempts to overwrite one of these variables, or if *behave* itself
+    tries to overwrite a user-set variable.
 
     You may use the "in" operator to test whether a certain value has been set
-    on the context, for example:
+    on the context, for example::
 
         "feature" in context
 
@@ -147,14 +157,13 @@ class Context(object):
     .. _`configuration file section names`: behave.html#configuration-files
     """
     # pylint: disable=too-many-instance-attributes
-    BEHAVE = "behave"
-    USER = "user"
+    LAYER_NAMES = ["testrun", "feature", "rule", "scenario"]
     FAIL_ON_CLEANUP_ERRORS = True
 
     def __init__(self, runner):
         self._runner = weakref.proxy(runner)
         self._config = runner.config
-        d = self._root = {
+        root_data = self._root = {
             "aborted": False,
             "failed": False,
             "config": self._config,
@@ -163,19 +172,72 @@ class Context(object):
             "@cleanups": [],    # -- REQUIRED-BY: before_all() hook
             "@layer": "testrun",
         }
-        self._stack = [d]
+        self._stack = [root_data]
         self._record = {}
         self._origin = {}
-        self._mode = self.BEHAVE
+        self._mode = ContextMode.BEHAVE
+
+        # -- MODEL ENTITY REFERENCES/SUPPORT:
+        # DISABLED: self.rule = None
+        # DISABLED: self.scenario = None
         self.feature = None
-        # -- RECHECK: If needed
         self.text = None
         self.table = None
-        self.stdout_capture = None
-        self.stderr_capture = None
-        self.log_capture = None
+
+        # -- RUNTIME SUPPORT:
         self.fail_on_cleanup_errors = self.FAIL_ON_CLEANUP_ERRORS
 
+    def abort(self, reason=None):
+        """Abort the test run.
+
+        This sets the :attr:`aborted` attribute to true.
+        Any test runner evaluates this attribute to abort a test run.
+
+        .. versionadded:: 1.2.7
+        """
+        if not self.aborted:
+            # -- ONLY ONCE:
+            print("ABORTED: {}".format(reason or ""), file=sys.stderr)
+            self._set_root_attribute("aborted", True)
+
+    def use_or_assign_param(self, name, value):
+        """Use an existing context parameter (aka: attribute) or
+        assign a value to new context parameter (if it does not exist yet).
+
+        :param name:   Context parameter name (as string)
+        :param value:  Parameter value for new parameter.
+        :return: Existing or newly created parameter.
+
+        .. versionadded:: 1.2.7
+        """
+        if name not in self:
+            # -- CASE: New, missing param -- Assign parameter-value.
+            setattr(self, name, value)
+            return value
+        # -- OTHERWISE: Use existing param
+        return getattr(self, name, None)
+
+
+    def use_or_create_param(self, name, factory_func, *args, **kwargs):
+        """Use an existing context parameter (aka: attribute) or
+        create a new parameter if it does not exist yet.
+
+        :param name:   Context parameter name (as string)
+        :param factory_func: Factory function, used if parameter is created.
+        :param args: Positional args for ``factory_func()`` on create.
+        :param kwargs: Named args for ``factory_func()`` on create.
+        :return: Existing or newly created parameter.
+
+        .. versionadded:: 1.2.7
+        """
+        if name not in self:
+            # -- CASE: New, missing param -- Create it.
+            param_value = factory_func(*args, **kwargs)
+            setattr(self, name, param_value)
+            return param_value
+        # -- OTHERWISE: Use existing param
+        return getattr(self, name, None)
+
     @staticmethod
     def ignore_cleanup_error(context, cleanup_func, exception):
         pass
@@ -210,7 +272,9 @@ class Context(object):
                 context.add_cleanup(cleanup_database, the_database)
         """
         # -- BEST-EFFORT ALGORITHM: Tries to perform all cleanups.
-        assert self._stack, "REQUIRE: Non-empty stack"
+        if not self._stack:
+            return
+
         current_layer = self._stack[0]
         cleanup_funcs = current_layer.get("@cleanups", [])
         on_cleanup_error = getattr(self, "on_cleanup_error",
@@ -226,40 +290,80 @@ class Context(object):
                 cleanup_errors.append(sys.exc_info())
                 on_cleanup_error(context, cleanup_func, e)
 
+        current_layer["@cleanups"] = []
         if self.fail_on_cleanup_errors and cleanup_errors:
             first_cleanup_erro_info = cleanup_errors[0]
             del cleanup_errors  # -- ENSURE: Release other exception frames.
             six.reraise(*first_cleanup_erro_info)
 
+    def _do_remaining_cleanups(self, capture_sink=None):
+        """
+        Performs any remaining cleanup operations.
 
-    def _push(self, layer_name=None):
+        This is done by removing the remaining layers from the context stack
+        and performs any pending cleanups.
+        """
+        # -- SINCE: behave v1.2.7
+        cleanup_errors = []
+        while len(self._stack) > 1:
+            try:
+                # -- PERFORM CLEANUPS:
+                self._pop(capture_sink=capture_sink)
+            except Exception:
+                cleanup_errors.append(sys.exc_info())
+
+        # -- FINALLY: Perform cleanups on last layer (if needed).
+        self._do_cleanups()
+        if self.fail_on_cleanup_errors and cleanup_errors:
+            first_cleanup_erro_info = cleanup_errors[0]
+            del cleanup_errors  # -- ENSURE: Release other exception frames.
+            six.reraise(*first_cleanup_erro_info)
+
+    def _push(self, layer=None):
         """Push a new layer on the context stack.
-        HINT: Use layer_name values: "scenario", "feature", "testrun".
+        HINT: Use layer values: "testrun", "feature", "rule, "scenario".
 
-        :param layer_name:   Layer name to use (or None).
+        :param layer:   Layer name to use (or None).
         """
         initial_data = {"@cleanups": []}
-        if layer_name:
-            initial_data["@layer"] = layer_name
+        if layer:
+            initial_data["@layer"] = layer
         self._stack.insert(0, initial_data)
 
-    def _pop(self):
-        """Pop the current layer from the context stack.
-        Performs any pending cleanups, registered for this layer.
+    def _pop(self, capture_sink=None):
         """
+        Pop the current layer from the context stack.
+        This performs any pending cleanups that are registered for this layer.
+        """
+        # -- PREPARED:
+        # if capture_sink is None:
+        #    capture_sink = self._runner.capture_sink
+
         try:
-            self._do_cleanups()
+            if self.config.should_capture_hooks():
+                layer = self._stack[0].get("@layer", None)
+                name = "{}.cleanup".format(layer or "")
+                show_on_success = any_hook.show_cleanup_on_success
+                with capture_output_to_sink(self._config,
+                                            capture_sink2=capture_sink,
+                                            name=name,
+                                            show_on_success=show_on_success):
+                    self._do_cleanups()
+            else:
+                self._do_cleanups()
         finally:
             # -- ENSURE: Layer is removed even if cleanup-errors occur.
             self._stack.pop(0)
 
+
+
     def _use_with_behave_mode(self):
         """Provides a context manager for using the context in BEHAVE mode."""
-        return use_context_with_mode(self, Context.BEHAVE)
+        return use_context_with_mode(self, ContextMode.BEHAVE)
 
     def use_with_user_mode(self):
         """Provides a context manager for using the context in USER mode."""
-        return use_context_with_mode(self, Context.USER)
+        return use_context_with_mode(self, ContextMode.USER)
 
     def user_mode(self):
         warnings.warn("Use 'use_with_user_mode()' instead",
@@ -286,11 +390,11 @@ class Context(object):
 
     def _emit_warning(self, attr, params):
         msg = ""
-        if self._mode is self.BEHAVE and self._origin[attr] is not self.BEHAVE:
+        if self._mode is ContextMode.BEHAVE and self._origin[attr] is not ContextMode.BEHAVE:
             msg = "behave runner is masking context attribute '%(attr)s' " \
                   "originally set in %(function)s (%(filename)s:%(line)s)"
-        elif self._mode is self.USER:
-            if self._origin[attr] is not self.USER:
+        elif self._mode is ContextMode.USER:
+            if self._origin[attr] is not ContextMode.USER:
                 msg = "user code is masking context attribute '%(attr)s' " \
                       "originally set by behave"
             elif self._config.verbose:
@@ -312,7 +416,11 @@ class Context(object):
 
     def __getattr__(self, attr):
         if attr[0] == "_":
-            return self.__dict__[attr]
+            try:
+                return self.__dict__[attr]
+            except KeyError:
+                raise AttributeError(attr)
+
         for frame in self._stack:
             if attr in frame:
                 return frame[attr]
@@ -377,7 +485,7 @@ class Context(object):
         :raises: AssertionError, if a step failure occurs.
         :raises: ValueError, if invoked without a feature context.
         """
-        assert isinstance(steps_text, six.text_type), "Steps must be unicode."
+        require_type(steps_text, six.text_type, message="Steps must be unicode")
         if not self.feature:
             raise ValueError("execute_steps() called outside of feature")
 
@@ -408,6 +516,20 @@ class Context(object):
             self.text = original_text
         return True
 
+    def _select_stack_frame_by_layer(self, layer):
+        """Select context stack frame by layer name.
+
+        :param layer:   Layer name (as string).
+        :return: Selected frame object (if any)
+        :raises: LookupError, if layer was not found.
+        """
+        for frame in self._stack:
+            frame_layer = frame.get("@layer", None)
+            if layer == frame_layer:
+                return frame
+        # -- OOPS, NOT FOUND:
+        raise LookupError("Context.stack: layer=%s not found" % layer)
+
     def add_cleanup(self, cleanup_func, *args, **kwargs):
         """Adds a cleanup function that is called when :meth:`Context._pop()`
         is called. This is intended for user-cleanups.
@@ -415,10 +537,23 @@ class Context(object):
         :param cleanup_func:    Callable function
         :param args:            Args for cleanup_func() call (optional).
         :param kwargs:          Kwargs for cleanup_func() call (optional).
+
+        .. note:: RESERVED :obj:`layer` : optional-string
+
+            The keyword argument ``layer="LAYER_NAME"`` can to be used to
+            assign the :obj:`cleanup_func` to specific a layer on the context stack
+            (instead of the current layer).
+
+            Known layer names are: "testrun", "feature", "rule", "scenario"
+
+        .. seealso:: :attr:`.Context.LAYER_NAMES`
         """
-        # MAYBE:
-        assert callable(cleanup_func), "REQUIRES: callable(cleanup_func)"
-        assert self._stack
+        # -- PRECONDITIONS:
+        require_callable(cleanup_func, message="REQUIRES: callable(cleanup_func)")
+        if not self._stack:
+            raise RuntimeError("Context._stack is empty")
+
+        layer_name = kwargs.pop("layer", None)
         if args or kwargs:
             def internal_cleanup_func():
                 cleanup_func(*args, **kwargs)
@@ -426,20 +561,37 @@ class Context(object):
             internal_cleanup_func = cleanup_func
 
         current_frame = self._stack[0]
+        if layer_name:
+            current_frame = self._select_stack_frame_by_layer(layer_name)
         if cleanup_func not in current_frame["@cleanups"]:
             # -- AVOID DUPLICATES:
             current_frame["@cleanups"].append(internal_cleanup_func)
 
+    @property
+    def captured(self):
+        return self._runner.captured
+
+    def attach(self, mime_type, data):
+        """Embeds data (e.g. a screenshot) in reports for all
+        formatters that support it, such as the JSON formatter.
+
+        :param mime_type:       MIME type of the binary data.
+        :param data:            Bytes-like object to embed.
+        """
+        is_compatible = lambda f: hasattr(f, "embedding")  # noqa: E731
+        for formatter in filter(is_compatible, self._runner.formatters):
+            formatter.embedding(mime_type, data)
+
 
 @contextlib.contextmanager
 def use_context_with_mode(context, mode):
-    """Switch context to BEHAVE or USER mode.
+    """Switch context to ContextMode.BEHAVE or ContextMode.USER mode.
     Provides a context manager for switching between the two context modes.
 
     .. sourcecode:: python
 
         context = Context()
-        with use_context_with_mode(context, Context.BEHAVE):
+        with use_context_with_mode(context, ContextMode.BEHAVE):
             ...     # Do something
         # -- POSTCONDITION: Original context._mode is restored.
 
@@ -447,7 +599,7 @@ def use_context_with_mode(context, mode):
     :param mode:     Mode to apply to context object.
     """
     # pylint: disable=protected-access
-    assert mode in (Context.BEHAVE, Context.USER)
+    assert mode in (ContextMode.BEHAVE, ContextMode.USER)
     current_mode = context._mode
     try:
         context._mode = mode
@@ -459,7 +611,7 @@ def use_context_with_mode(context, mode):
 
 
 @contextlib.contextmanager
-def scoped_context_layer(context, layer_name=None):
+def scoped_context_layer(context, layer=None):
     """Provides context manager for context layer (push/do-something/pop cycle).
 
     .. code-block::
@@ -469,7 +621,7 @@ def scoped_context_layer(context, layer_name=None):
     """
     # pylint: disable=protected-access
     try:
-        context._push(layer_name)
+        context._push(layer)
         yield context
     finally:
         context._pop()
@@ -480,10 +632,16 @@ def path_getrootdir(path):
     Extract rootdir from path in a platform independent way.
 
     POSIX-PATH EXAMPLE:
+
+    .. code-block:: python
+
         rootdir = path_getrootdir("/foo/bar/one.feature")
         assert rootdir == "/"
 
     WINDOWS-PATH EXAMPLE:
+
+    .. code-block:: python
+
         rootdir = path_getrootdir("D:\\foo\\bar\\one.feature")
         assert rootdir == r"D:\"
     """
@@ -496,8 +654,7 @@ def path_getrootdir(path):
 
 
 class ModelRunner(object):
-    """
-    Test runner for a behave model (features).
+    """Test runner for a behave model (features).
     Provides the core functionality of a test runner and
     the functional API needed by model elements.
 
@@ -506,6 +663,14 @@ class ModelRunner(object):
           This is set to true when the user aborts a test run
           (:exc:`KeyboardInterrupt` exception). Initially: False.
           Stored as derived attribute in :attr:`Context.aborted`.
+
+    .. attribute:: captured
+
+        If any output capture is enabled, provides access to a
+        :class:`~behave.capture.Captured` object that contains a snapshot
+        of all captured data (stdout/stderr/log).
+
+        .. versionadded:: 1.3.0
     """
     # pylint: disable=too-many-instance-attributes
 
@@ -514,80 +679,216 @@ class ModelRunner(object):
         self.features = features or []
         self.hooks = {}
         self.formatters = []
-        self.undefined_steps = []
+        self._undefined_steps = []
         self.step_registry = step_registry
         self.capture_controller = CaptureController(config)
+        self._captured = ManyCaptured()
+        self.capture_sink = CaptureSinkAsCollector(self._captured,
+                                                   store_on_success=True)
 
         self.context = None
         self.feature = None
         self.hook_failures = 0
 
-    # @property
-    def _get_aborted(self):
-        value = False
+    @property
+    def undefined_steps(self):
+        return self._undefined_steps
+
+    @property
+    def aborted(self):
+        """Indicates that test run is aborted by the user or system."""
         if self.context:
-            value = self.context.aborted
-        return value
+            return self.context.aborted
+        # -- OTHERWISE
+        return False
 
-    # @aborted.setter
-    def _set_aborted(self, value):
+    @aborted.setter
+    def aborted(self, value):
+        """Mark the test run as aborted."""
         # pylint: disable=protected-access
-        assert self.context, "REQUIRE: context, but context=%r" % self.context
-        self.context._set_root_attribute("aborted", bool(value))
+        require_not_none(self.context, name="ModelRunner.context")
+        if self.context:
+            self.context._set_root_attribute("aborted", bool(value))
 
-    aborted = property(_get_aborted, _set_aborted,
-                       doc="Indicates that test run is aborted by the user.")
+    # DISABLED: aborted = property(_get_aborted, _set_aborted, doc="...")
 
-    def run_hook(self, name, context, *args):
-        if not self.config.dry_run and (name in self.hooks):
-            try:
-                with context.use_with_user_mode():
-                    self.hooks[name](context, *args)
-            # except KeyboardInterrupt:
-            #     self.aborted = True
-            #     if name not in ("before_all", "after_all"):
-            #         raise
-            except Exception as e:  # pylint: disable=broad-except
-                # -- HANDLE HOOK ERRORS:
-                use_traceback = False
-                if self.config.verbose:
-                    use_traceback = True
-                    ExceptionUtil.set_traceback(e)
-                extra = u""
-                if "tag" in name:
-                    extra = "(tag=%s)" % args[0]
+    def abort(self, reason=None):
+        """Abort the test run.
 
-                error_text = ExceptionUtil.describe(e, use_traceback).rstrip()
-                error_message = u"HOOK-ERROR in %s%s: %s" % (name, extra, error_text)
-                print(error_message)
-                self.hook_failures += 1
-                if "tag" in name:
-                    # -- SCENARIO or FEATURE
-                    statement = getattr(context, "scenario", context.feature)
-                elif "all" in name:
-                    # -- ABORT EXECUTION: For before_all/after_all
-                    self.aborted = True
-                    statement = None
+        .. versionadded:: 1.2.7
+        """
+        if self.context is None:
+            return  # -- GRACEFULLY IGNORED.
+
+        # -- NORMAL CASE:
+        # SIMILAR TO: self.aborted = True
+        self.context.abort(reason=reason)
+
+    def should_run_hook(self, hook_name):
+        # -- SINCE: behave v1.2.7
+        return not self.config.dry_run and (hook_name in self.hooks)
+
+    # OLD: def run_hook(self, hook_name, context, *args):
+    def run_hook(self, hook_name, *args):
+        if not self.should_run_hook(hook_name):
+            # -- SHORTCUT: No need to run-hook -- HOOK_PASSED (gracefully)
+            return True
+        if ("all" in hook_name and len(args) > 0) or len(args) > 1:
+            # -- DEPRECATED: context parameter was provided.
+            # SINCE: behave v1.2.7
+            warnings.warn("Avoid 'context' in run_hook(hook_name, context, ...)",
+                          DeprecationWarning)
+            args = args[1:]
+
+        def select_current_statement_for_tag(ctx):
+            for statement_name in ("scenario", "rule", "feature"):
+                this_statement = getattr(ctx, statement_name, None)
+                if this_statement:
+                    return this_statement
+            return None
+
+        def select_current_statement_for(hook_name, args, ctx):
+            this_statement = None
+            if args and "all" not in hook_name:
+                this_statement = args[0]
+                if "tag" in hook_name:
+                    this_statement = select_current_statement_for_tag(ctx)
+            return this_statement
+
+        ctx = self.context
+        # statement = None
+        # if args and "all" not in hook_name:
+        #     statement = args[0]
+        #     if "tag" in hook_name:
+        #         # -- SCENARIO or RULE or FEATURE
+        #         statement = select_current_statement_for_tag(ctx)
+
+        raise_exception_enabled = False
+        try:
+            with ctx.use_with_user_mode():
+                self.hooks[hook_name](ctx, *args)
+                return True  # -- HOOK_PASSED
+        except (KeyboardInterrupt, SystemExit) as e:
+            e_type = type(e).__name__
+            message = "HOOK-ERROR in {}: {}".format(hook_name, e_type)
+            self.abort(reason=message)
+            self.hook_failures += 1
+            statement = select_current_statement_for(hook_name, args, ctx)
+            if statement:
+                statement.hook_failed = True
+            if raise_exception_enabled and hook_name not in ("before_all", "after_all"):
+                raise
+        except Exception as e:  # pylint: disable=broad-except
+            # -- HANDLE HOOK ERRORS:
+            use_traceback = False
+            if self.config.verbose:
+                use_traceback = True
+                ExceptionUtil.set_traceback(e)
+            extra = u""
+            if "tag" in hook_name:
+                extra = "(tag=%s)" % args[0]
+
+            error_text = ExceptionUtil.describe(e, use_traceback).rstrip()
+            error_message = u"HOOK-ERROR in %s%s: %s" % (hook_name, extra, error_text)
+            print(error_message)  # -- MAYBE: Use stderr
+            self.hook_failures += 1
+            # if "tag" in hook_name:
+            #    # -- SCENARIO or FEATURE
+            #    statement = getattr(ctx, "scenario", ctx.feature)
+            if hook_name == "before_all":
+                # -- ABORT EXECUTION: For before_all/after_all
+                self.abort(reason="HOOK-ERROR in hook=%s" % hook_name)
+                # statement = None
+                pass
+            # else:
+            #    # -- CASE: feature, scenario, step
+            #    statement = args[0]
+
+            statement = select_current_statement_for(hook_name, args, ctx)
+            if statement:
+                # -- CASE: feature, scenario, step
+                statement.hook_failed = True
+                if statement.error_message:
+                    # -- NOTE: One exception/failure is already stored.
+                    #    Append only error message.
+                    statement.error_message += u"\n"+ error_message
                 else:
-                    # -- CASE: feature, scenario, step
-                    statement = args[0]
+                    # -- FIRST EXCEPTION/FAILURE:
+                    statement.store_exception_context(e)
+                    statement.error_message = error_message
+            return False  # -- HOOK_FAILED
 
-                if statement:
-                    # -- CASE: feature, scenario, step
-                    statement.hook_failed = True
-                    if statement.error_message:
-                        # -- NOTE: One exception/failure is already stored.
-                        #    Append only error message.
-                        statement.error_message += u"\n"+ error_message
-                    else:
-                        # -- FIRST EXCEPTION/FAILURE:
-                        statement.store_exception_context(e)
-                        statement.error_message = error_message
+    def should_capture_hook(self, hook_name):
+        hook_func = self.hooks.get(hook_name, None)
+        if not (hook_func and self.config.should_capture_hooks()):
+            return False
 
-    def setup_capture(self):
-        if not self.context:
-            self.context = Context(self)
-        self.capture_controller.setup_capture(self.context)
+        # -- SUPPORT: Disable the capture-hooks mode on a hook.
+        # EXAMPLE: before_scenario.capture = False
+        capture_hook_enabled = getattr(hook_func, "capture", True)
+        return bool(capture_hook_enabled)
+
+    def run_hook_with_capture(self, hook_name, *args, **kwargs):
+        # -- SINCE: behave v1.2.7
+        if not self.should_run_hook(hook_name):
+            # -- SHORTCUT: No need to run MISSING_HOOK -- PASSED
+            return True
+        if not self.should_capture_hook(hook_name):
+            # -- SHORTCUT: NO-CAPTURE CASE -- no need to capture
+            return self.run_hook(hook_name, *args)
+
+        capture_sink = kwargs.pop("capture_sink", self.capture_sink)
+        if kwargs:
+            raise ValueError("UNSUPPORTED: {}".format(", ".join(kwargs.keys())))
+
+        # -- CAPTURE: run-hook
+        this_hook = self.hooks[hook_name]
+        show_on_success = getattr(this_hook, "show_capture_on_success",
+                                  any_hook.show_capture_on_success)
+        capture_kwargs = dict(capture_sink2=capture_sink, name=hook_name,
+                              show_on_success=show_on_success)
+        hook_passed = False
+        with capture_output_to_sink(self.config, **capture_kwargs) as controller:
+            hook_passed = self.run_hook(hook_name, *args)
+            controller.result_failed = not hook_passed
+        return hook_passed
+
+    def run_hook_tags(self, hook_name, tags):
+        if hook_name not in ("before_tag", "after_tag"):
+            msg = "hook_name={} (expected: before_tag, after_tag".format(hook_name)
+            raise ValueError(msg)
+
+        failed_count = 0
+        for tag in tags:
+            hook_passed = self.run_hook(hook_name, tag)
+            if not hook_passed:
+                failed_count += 1
+        return failed_count == 0  # -- HINT: ALL_HOOKS_PASSED
+
+    def run_hook_tags_with_capture(self, hook_name, tags, capture_sink=None):
+        # -- SINCE: behave v1.2.7
+        if not self.should_run_hook(hook_name):
+            # -- SHORTCUT: No need to run-hook MISSING_HOOK -- PASSED
+            return True
+        if not self.should_capture_hook(hook_name):
+            # -- SHORTCUT: NO-CAPTURE CASE -- no need to capture
+            return self.run_hook_tags(hook_name, tags)
+
+        # -- CAPTURE: run-hook
+        this_hook = self.hooks[hook_name]
+        show_on_success = getattr(this_hook, "show_capture_on_success",
+                                  any_hook.show_capture_on_success)
+        capture_kwargs = dict(capture_sink2=capture_sink, name=hook_name,
+                              show_on_success=show_on_success)
+        hook_passed = False
+        with capture_output_to_sink(self.config, **capture_kwargs) as controller:
+            hook_passed = self.run_hook_tags(hook_name, tags)
+            controller.result_failed = not hook_passed
+        return hook_passed
+
+    def setup_capture(self, name=None):
+        require_not_type(name, Context)
+        self.capture_controller.setup_capture(name=name)
 
     def start_capture(self):
         self.capture_controller.start_capture()
@@ -598,6 +899,18 @@ class ModelRunner(object):
     def teardown_capture(self):
         self.capture_controller.teardown_capture()
 
+    @property
+    def captured(self):
+        """Return the current state of the captured output/logging
+        (as captured object).
+        """
+        return self.capture_controller.captured
+        # -- DISABLED:
+        # if self.capture_controller.has_output():
+        #     captured = self.capture_controller.make_captured_delta()
+        #     self._captured.add_captured(captured)
+        # return self._captured
+
     def run_model(self, features=None):
         # pylint: disable=too-many-branches
         if not self.context:
@@ -608,10 +921,11 @@ class ModelRunner(object):
             features = self.features
 
         # -- ENSURE: context.execute_steps() works in weird cases (hooks, ...)
-        context = self.context
         self.hook_failures = 0
-        self.setup_capture()
-        self.run_hook("before_all", context)
+        # -- DISABLED:
+        # self.setup_capture()
+        # self.run_hook_with_capture("before_all")
+        self.run_hook("before_all")
 
         run_feature = not self.aborted
         failed_count = 0
@@ -630,7 +944,7 @@ class ModelRunner(object):
                             # -- FAIL-EARLY: After first failure.
                             run_feature = False
                 except KeyboardInterrupt:
-                    self.aborted = True
+                    self.abort(reason="KeyboardInterrupt")
                     failed_count += 1
                     run_feature = False
 
@@ -642,12 +956,19 @@ class ModelRunner(object):
         # -- AFTER-ALL:
         # pylint: disable=protected-access, broad-except
         cleanups_failed = False
-        self.run_hook("after_all", self.context)
+        self.run_hook_with_capture("after_all")
         try:
-            self.context._do_cleanups()   # Without dropping the last context layer.
+            # -- PERFORM CLEANUPS: Without dropping the last context layer.
+            self.context._do_remaining_cleanups()
         except Exception:
             cleanups_failed = True
 
+        # -- DUPLICATES CAPTURE-OUTPUT:
+        #   Using capture_controller instead captured/capture_sink
+        # if self.captured.has_output():
+        #    captured_output = self.captured.make_report()
+        #    print(captured_output)
+
         if self.aborted:
             print("\nABORTED: By user.")
         for formatter in self.formatters:
@@ -655,10 +976,12 @@ class ModelRunner(object):
         for reporter in self.config.reporters:
             reporter.end()
 
+        # -- MAYBE: BAD STEP-DEFINITIONS: Unused BAD STEPS should not cause FAILURE.
+        # bad_step_definitions = self.step_registry.error_handler.bad_step_definitions
         failed = ((failed_count > 0) or self.aborted or (self.hook_failures > 0)
                   or (len(self.undefined_steps) > undefined_steps_initial_size)
                   or cleanups_failed)
-                  # XXX-MAYBE: or context.failed)
+                  # -- MAYBE: or context.failed)
         return failed
 
     def run(self):
@@ -670,20 +993,20 @@ class ModelRunner(object):
 
 
 class Runner(ModelRunner):
-    """
-    Standard test runner for behave:
+    """Standard test runner for behave:
 
       * setup paths
       * loads environment hooks
       * loads step definitions
       * select feature files, parses them and creates model (elements)
     """
+    DEFAULT_DIRECTORY = "features"
+
     def __init__(self, config):
         super(Runner, self).__init__(config)
         self.path_manager = PathManager()
         self.base_dir = None
 
-
     def setup_paths(self):
         # pylint: disable=too-many-branches, too-many-statements
         if self.config.paths:
@@ -710,8 +1033,8 @@ class Runner(ModelRunner):
                 base_dir = os.path.dirname(base_dir)
         else:
             if self.config.verbose:
-                print('Using default path "./features"')
-            base_dir = os.path.abspath("features")
+                print('Using default path "{}"'.format(self.DEFAULT_DIRECTORY))
+            base_dir = os.path.abspath(self.DEFAULT_DIRECTORY)
 
         # Get the root. This is not guaranteed to be "/" because Windows.
         root_dir = path_getrootdir(base_dir)
@@ -748,7 +1071,7 @@ class Runner(ModelRunner):
         base_dir = new_base_dir
         self.config.base_dir = base_dir
 
-        for dirpath, dirnames, filenames in os.walk(base_dir):
+        for dirpath, dirnames, filenames in os.walk(base_dir, followlinks=True):
             if [fn for fn in filenames if fn.endswith(".feature")]:
                 break
         else:
@@ -778,8 +1101,9 @@ class Runner(ModelRunner):
         context.config.setup_logging()
 
     def load_hooks(self, filename=None):
+        base_dir = self.base_dir or self.DEFAULT_DIRECTORY
         filename = filename or self.config.environment_file
-        hooks_path = os.path.join(self.base_dir, filename)
+        hooks_path = os.path.join(base_dir, filename)
         if os.path.exists(hooks_path):
             exec_file(hooks_path, self.hooks)
 
@@ -787,12 +1111,22 @@ class Runner(ModelRunner):
             self.hooks["before_all"] = self.before_all_default_hook
 
     def load_step_definitions(self, extra_step_paths=None):
+        """
+        Allow steps to import other stuff from the steps dir
+        NOTE: Default matcher can be overridden in "environment.py" hook.
+        Collect directories containing step files
+        """
         if extra_step_paths is None:
             extra_step_paths = []
-        # -- Allow steps to import other stuff from the steps dir
-        # NOTE: Default matcher can be overridden in "environment.py" hook.
+
         steps_dir = os.path.join(self.base_dir, self.config.steps_dir)
-        step_paths = [steps_dir] + list(extra_step_paths)
+        step_paths = set()
+
+        for root, _, files in os.walk(steps_dir):
+            step_paths.add(root)
+
+        step_paths = list(step_paths) + list(extra_step_paths)
+
         load_step_modules(step_paths)
 
     def feature_locations(self):
@@ -810,7 +1144,7 @@ class Runner(ModelRunner):
 
         # -- ENSURE: context.execute_steps() works in weird cases (hooks, ...)
         # self.setup_capture()
-        # self.run_hook("before_all", self.context)
+        # self.run_hook("before_all")
 
         # -- STEP: Parse all feature files (by using their file location).
         feature_locations = [filename for filename in self.feature_locations()
@@ -822,3 +1156,10 @@ class Runner(ModelRunner):
         stream_openers = self.config.outputs
         self.formatters = make_formatters(self.config, stream_openers)
         return self.run_model()
+
+
+# -----------------------------------------------------------------------------
+# REGISTER RUNNER-CLASSES:
+# -----------------------------------------------------------------------------
+ITestRunner.register(ModelRunner)
+ITestRunner.register(Runner)
